@@ -5,44 +5,42 @@ import logging
 import time
 from concurrent.futures.process import ProcessPoolExecutor
 from typing import Dict, List, Optional, Set, Tuple
-
 from blspy import AugSchemeMPL, G1Element
 from chiabip158 import PyBIP158
 
 from chia.consensus.block_record import BlockRecord
 from chia.consensus.constants import ConsensusConstants
-from chia.consensus.cost_calculator import CostResult, calculate_cost_of_program
-from chia.full_node.bundle_tools import best_solution_program
+from chia.consensus.cost_calculator import NPCResult, calculate_cost_of_program
+from chia.full_node.bundle_tools import simple_solution_generator
 from chia.full_node.coin_store import CoinStore
 from chia.full_node.mempool import Mempool
-from chia.full_node.mempool_check_conditions import mempool_check_conditions_dict
+from chia.full_node.mempool_check_conditions import mempool_check_conditions_dict, get_name_puzzle_conditions
 from chia.types.blockchain_format.coin import Coin
 from chia.types.blockchain_format.sized_bytes import bytes32
 from chia.types.coin_record import CoinRecord
 from chia.types.condition_opcodes import ConditionOpcode
-from chia.types.condition_var_pair import ConditionVarPair
-from chia.types.full_block import additions_for_npc
+from chia.types.condition_with_args import ConditionWithArgs
 from chia.types.mempool_inclusion_status import MempoolInclusionStatus
 from chia.types.mempool_item import MempoolItem
 from chia.types.spend_bundle import SpendBundle
 from chia.util.clvm import int_from_bytes
-from chia.util.condition_tools import pkm_pairs_for_conditions_dict
+from chia.util.condition_tools import (
+    pkm_pairs_for_conditions_dict,
+    coin_announcements_names_for_npc,
+    puzzle_announcements_names_for_npc,
+)
 from chia.util.errors import Err
+from chia.util.generator_tools import additions_for_npc
 from chia.util.ints import uint32, uint64
-from chia.util.streamable import dataclass_from_dict, recurse_jsonify
+from chia.util.streamable import recurse_jsonify
 
 log = logging.getLogger(__name__)
 
 
-def validate_transaction_multiprocess(
-    constants_dict: Dict,
-    spend_bundle_bytes: bytes,
-) -> bytes:
-    constants: ConsensusConstants = dataclass_from_dict(ConsensusConstants, constants_dict)
-    # Calculate the cost and fees
-    program = best_solution_program(SpendBundle.from_bytes(spend_bundle_bytes))
+def get_npc_multiprocess(spend_bundle_bytes: bytes, max_cost: int) -> bytes:
+    program = simple_solution_generator(SpendBundle.from_bytes(spend_bundle_bytes))
     # npc contains names of the coins removed, puzzle_hashes and their spend conditions
-    return bytes(calculate_cost_of_program(program, constants.CLVM_COST_RATIO_CONSTANT, True))
+    return bytes(get_name_puzzle_conditions(program, max_cost, True))
 
 
 class MempoolManager:
@@ -51,25 +49,22 @@ class MempoolManager:
         self.constants_json = recurse_jsonify(dataclasses.asdict(self.constants))
 
         # Transactions that were unable to enter mempool, used for retry. (they were invalid)
-        self.potential_txs: Dict[bytes32, Tuple[SpendBundle, CostResult, bytes32]] = {}
+        self.potential_txs: Dict[bytes32, MempoolItem] = {}
         # Keep track of seen spend_bundles
         self.seen_bundle_hashes: Dict[bytes32, bytes32] = {}
 
         self.coin_store = coin_store
 
-        tx_per_sec = self.constants.TX_PER_SEC
-        sec_per_block = self.constants.SUB_SLOT_TIME_TARGET // self.constants.SLOT_BLOCKS_TARGET
-        block_buffer_count = self.constants.MEMPOOL_BLOCK_BUFFER
-
-        # MEMPOOL_SIZE = 60000
-        self.mempool_size = int(tx_per_sec * sec_per_block * block_buffer_count)
-        self.potential_cache_size = 300
+        self.limit_factor = 0.5
+        self.mempool_max_total_cost = int(self.constants.MAX_BLOCK_COST_CLVM * self.constants.MEMPOOL_BLOCK_BUFFER)
+        self.potential_cache_max_total_cost = int(self.constants.MAX_BLOCK_COST_CLVM * 5)
+        self.potential_cache_cost: int = 0
         self.seen_cache_size = 10000
         self.pool = ProcessPoolExecutor(max_workers=1)
 
         # The mempool will correspond to a certain peak
         self.peak: Optional[BlockRecord] = None
-        self.mempool: Mempool = Mempool.create(self.mempool_size)
+        self.mempool: Mempool = Mempool(self.mempool_max_total_cost)
 
     def shut_down(self):
         self.pool.shutdown(wait=True)
@@ -84,7 +79,7 @@ class MempoolManager:
         if (
             self.peak is None
             or self.peak.header_hash != peak_header_hash
-            or self.peak.height <= self.constants.INITIAL_FREEZE_PERIOD
+            or int(time.time()) <= self.constants.INITIAL_FREEZE_END_TIMESTAMP
         ):
             return None
 
@@ -93,21 +88,34 @@ class MempoolManager:
         spend_bundles: List[SpendBundle] = []
         removals = []
         additions = []
+        broke_from_inner_loop = False
+        log.info(f"Starting to make block, max cost: {self.constants.MAX_BLOCK_COST_CLVM}")
         for dic in self.mempool.sorted_spends.values():
+            if broke_from_inner_loop:
+                break
             for item in dic.values():
+                log.info(f"Cumulative cost: {cost_sum}")
                 if (
-                    item.cost_result.cost + cost_sum <= self.constants.MAX_BLOCK_COST_CLVM
+                    item.cost + cost_sum <= self.limit_factor * self.constants.MAX_BLOCK_COST_CLVM
                     and item.fee + fee_sum <= self.constants.MAX_COIN_AMOUNT
                 ):
                     spend_bundles.append(item.spend_bundle)
-                    cost_sum += item.cost_result.cost
+                    cost_sum += item.cost
                     fee_sum += item.fee
                     removals.extend(item.removals)
                     additions.extend(item.additions)
                 else:
+                    broke_from_inner_loop = True
                     break
         if len(spend_bundles) > 0:
-            return SpendBundle.aggregate(spend_bundles), additions, removals
+            log.info(
+                f"Cumulative cost of block (real cost should be less) {cost_sum}. Proportion "
+                f"full: {cost_sum / self.constants.MAX_BLOCK_COST_CLVM}"
+            )
+            agg = SpendBundle.aggregate(spend_bundles)
+            assert set(agg.additions()) == set(additions)
+            assert set(agg.removals()) == set(removals)
+            return agg, additions, removals
         else:
             return None
 
@@ -130,7 +138,7 @@ class MempoolManager:
         if cost == 0:
             return False
         fees_per_cost = fees / cost
-        if not self.mempool.at_full_capacity() or fees_per_cost >= self.mempool.get_min_fee_rate():
+        if not self.mempool.at_full_capacity(cost) or fees_per_cost > self.mempool.get_min_fee_rate(cost):
             return True
         return False
 
@@ -148,23 +156,57 @@ class MempoolManager:
         if bundle_hash in self.seen_bundle_hashes:
             self.seen_bundle_hashes.pop(bundle_hash)
 
-    async def pre_validate_spendbundle(self, new_spend: SpendBundle) -> CostResult:
+    @staticmethod
+    def get_min_fee_increase() -> int:
+        # 0.00001 XCH
+        return 10000000
+
+    def can_replace(
+        self,
+        conflicting_items: Dict[bytes32, MempoolItem],
+        removals: Dict[bytes32, CoinRecord],
+        fees: uint64,
+        fees_per_cost: float,
+    ) -> bool:
+        conflicting_fees = 0
+        conflicting_cost = 0
+        for item in conflicting_items.values():
+            conflicting_fees += item.fee
+            conflicting_cost += item.cost
+
+            # All coins spent in all conflicting items must also be spent in
+            # the new item
+            for coin in item.removals:
+                if coin.name() not in removals:
+                    return False
+
+        # New item must have higher fee per cost
+        if fees_per_cost <= conflicting_fees / conflicting_cost:
+            return False
+
+        # New item must increase the total fee at least by a certain amount
+        if fees < conflicting_fees + self.get_min_fee_increase():
+            return False
+
+        return True
+
+    async def pre_validate_spendbundle(self, new_spend: SpendBundle) -> NPCResult:
         """
         Errors are included within the cached_result.
         This runs in another process so we don't block the main thread
         """
         start_time = time.time()
         cached_result_bytes = await asyncio.get_running_loop().run_in_executor(
-            self.pool, validate_transaction_multiprocess, self.constants_json, bytes(new_spend)
+            self.pool, get_npc_multiprocess, bytes(new_spend), self.constants.MAX_BLOCK_COST_CLVM
         )
         end_time = time.time()
         log.info(f"It took {end_time - start_time} to pre validate transaction")
-        return CostResult.from_bytes(cached_result_bytes)
+        return NPCResult.from_bytes(cached_result_bytes)
 
     async def add_spendbundle(
         self,
         new_spend: SpendBundle,
-        cost_result: CostResult,
+        npc_result: NPCResult,
         spend_name: bytes32,
         validate_signature=True,
     ) -> Tuple[Optional[uint64], MempoolInclusionStatus, Optional[Err]]:
@@ -176,16 +218,17 @@ class MempoolManager:
         if self.peak is None:
             return None, MempoolInclusionStatus.FAILED, Err.MEMPOOL_NOT_INITIALIZED
 
-        npc_list = cost_result.npc_list
-        cost = cost_result.cost
+        npc_list = npc_result.npc_list
+        program = simple_solution_generator(new_spend).program
+        cost = calculate_cost_of_program(program, npc_result, self.constants.COST_PER_BYTE)
 
         log.debug(f"Cost: {cost}")
 
-        if cost > self.constants.MAX_BLOCK_COST_CLVM:
+        if cost > int(self.limit_factor * self.constants.MAX_BLOCK_COST_CLVM):
             return None, MempoolInclusionStatus.FAILED, Err.BLOCK_COST_EXCEEDS_MAX
 
-        if cost_result.error is not None:
-            return None, MempoolInclusionStatus.FAILED, Err(cost_result.error)
+        if npc_result.error is not None:
+            return None, MempoolInclusionStatus.FAILED, Err(npc_result.error)
         # build removal list
         removal_names: List[bytes32] = new_spend.removal_names()
 
@@ -231,19 +274,23 @@ class MempoolManager:
             elif name in additions_dict:
                 removal_coin = additions_dict[name]
                 # TODO(straya): what timestamp to use here?
+                assert self.peak.timestamp is not None
                 removal_record = CoinRecord(
                     removal_coin,
                     uint32(self.peak.height + 1),  # In mempool, so will be included in next height
                     uint32(0),
                     False,
                     False,
-                    uint64(int(time.time())),
+                    uint64(self.peak.timestamp + 1),
                 )
 
             assert removal_record is not None
             removal_amount = uint64(removal_amount + removal_record.coin.amount)
             removal_record_dict[name] = removal_record
             removal_coin_dict[name] = removal_record.coin
+
+        removals: List[Coin] = [coin for coin in removal_coin_dict.values()]
+
         if unknown_unspent_error:
             return None, MempoolInclusionStatus.FAILED, Err.UNKNOWN_UNSPENT
 
@@ -251,14 +298,16 @@ class MempoolManager:
             print(addition_amount, removal_amount)
             return None, MempoolInclusionStatus.FAILED, Err.MINTING_COIN
 
-        fees = removal_amount - addition_amount
+        fees = uint64(removal_amount - addition_amount)
         assert_fee_sum: uint64 = uint64(0)
 
         for npc in npc_list:
             if ConditionOpcode.RESERVE_FEE in npc.condition_dict:
-                fee_list: List[ConditionVarPair] = npc.condition_dict[ConditionOpcode.RESERVE_FEE]
+                fee_list: List[ConditionWithArgs] = npc.condition_dict[ConditionOpcode.RESERVE_FEE]
                 for cvp in fee_list:
                     fee = int_from_bytes(cvp.vars[0])
+                    if fee < 0:
+                        return None, MempoolInclusionStatus.FAILED, Err.RESERVE_FEE_CONDITION_FAILED
                     assert_fee_sum = assert_fee_sum + fee
         if fees < assert_fee_sum:
             return (
@@ -272,10 +321,10 @@ class MempoolManager:
 
         fees_per_cost: float = fees / cost
         # If pool is at capacity check the fee, if not then accept even without the fee
-        if self.mempool.at_full_capacity():
+        if self.mempool.at_full_capacity(cost):
             if fees == 0:
                 return None, MempoolInclusionStatus.FAILED, Err.INVALID_FEE_LOW_FEE
-            if fees_per_cost < self.mempool.get_min_fee_rate():
+            if fees_per_cost <= self.mempool.get_min_fee_rate(cost):
                 return None, MempoolInclusionStatus.FAILED, Err.INVALID_FEE_LOW_FEE
         # Check removals against UnspentDB + DiffStore + Mempool + SpendBundle
         # Use this information later when constructing a block
@@ -287,14 +336,14 @@ class MempoolManager:
             for conflicting in conflicts:
                 sb: MempoolItem = self.mempool.removals[conflicting.name()]
                 conflicting_pool_items[sb.name] = sb
-            for item in conflicting_pool_items.values():
-                if item.fee_per_cost >= fees_per_cost:
-                    self.add_to_potential_tx_set(new_spend, spend_name, cost_result)
-                    return (
-                        uint64(cost),
-                        MempoolInclusionStatus.PENDING,
-                        Err.MEMPOOL_CONFLICT,
-                    )
+            if not self.can_replace(conflicting_pool_items, removal_record_dict, fees, fees_per_cost):
+                potential = MempoolItem(new_spend, uint64(fees), npc_result, cost, spend_name, additions, removals)
+                self.add_to_potential_tx_set(potential)
+                return (
+                    uint64(cost),
+                    MempoolInclusionStatus.PENDING,
+                    Err.MEMPOOL_CONFLICT,
+                )
 
         elif fail_reason:
             return None, MempoolInclusionStatus.FAILED, fail_reason
@@ -306,6 +355,8 @@ class MempoolManager:
         pks: List[G1Element] = []
         msgs: List[bytes32] = []
         error: Optional[Err] = None
+        coin_announcements_in_spend: Set[bytes32] = coin_announcements_names_for_npc(npc_list)
+        puzzle_announcements_in_spend: Set[bytes32] = puzzle_announcements_names_for_npc(npc_list)
         for npc in npc_list:
             coin_record: CoinRecord = removal_record_dict[npc.coin_name]
             # Check that the revealed removal puzzles actually match the puzzle hash
@@ -317,16 +368,27 @@ class MempoolManager:
             chialisp_height = (
                 self.peak.prev_transaction_block_height if not self.peak.is_transaction_block else self.peak.height
             )
-            error = mempool_check_conditions_dict(coin_record, new_spend, npc.condition_dict, uint32(chialisp_height))
+            assert self.peak.timestamp is not None
+            error = mempool_check_conditions_dict(
+                coin_record,
+                coin_announcements_in_spend,
+                puzzle_announcements_in_spend,
+                npc.condition_dict,
+                uint32(chialisp_height),
+                self.peak.timestamp,
+            )
 
             if error:
-                if error is Err.ASSERT_HEIGHT_NOW_EXCEEDS_FAILED or error is Err.ASSERT_HEIGHT_AGE_EXCEEDS_FAILED:
-                    self.add_to_potential_tx_set(new_spend, spend_name, cost_result)
+                if error is Err.ASSERT_HEIGHT_ABSOLUTE_FAILED or error is Err.ASSERT_HEIGHT_RELATIVE_FAILED:
+                    potential = MempoolItem(new_spend, uint64(fees), npc_result, cost, spend_name, additions, removals)
+                    self.add_to_potential_tx_set(potential)
                     return uint64(cost), MempoolInclusionStatus.PENDING, error
                 break
 
             if validate_signature:
-                for pk, message in pkm_pairs_for_conditions_dict(npc.condition_dict, npc.coin_name):
+                for pk, message in pkm_pairs_for_conditions_dict(
+                    npc.condition_dict, npc.coin_name, self.constants.AGG_SIG_ME_ADDITIONAL_DATA
+                ):
                     pks.append(pk)
                     msgs.append(message)
         if error:
@@ -341,10 +403,9 @@ class MempoolManager:
         if fail_reason:
             mempool_item: MempoolItem
             for mempool_item in conflicting_pool_items.values():
-                self.mempool.remove_spend(mempool_item)
+                self.mempool.remove_from_pool(mempool_item)
 
-        removals: List[Coin] = [coin for coin in removal_coin_dict.values()]
-        new_item = MempoolItem(new_spend, uint64(fees), cost_result, spend_name, additions, removals)
+        new_item = MempoolItem(new_spend, uint64(fees), npc_result, cost, spend_name, additions, removals)
         self.mempool.add_to_pool(new_item, additions, removal_coin_dict)
         log.info(f"add_spendbundle took {time.time() - start_time} seconds")
         return uint64(cost), MempoolInclusionStatus.SUCCESS, None
@@ -373,15 +434,20 @@ class MempoolManager:
         # 5. If coins can be spent return list of unspents as we see them in local storage
         return None, []
 
-    def add_to_potential_tx_set(self, spend: SpendBundle, spend_name: bytes32, cost_result: CostResult):
+    def add_to_potential_tx_set(self, item: MempoolItem):
         """
         Adds SpendBundles that have failed to be added to the pool in potential tx set.
         This is later used to retry to add them.
         """
-        self.potential_txs[spend_name] = spend, cost_result, spend_name
+        if item.spend_bundle_name in self.potential_txs:
+            return
 
-        while len(self.potential_txs) > self.potential_cache_size:
+        self.potential_txs[item.spend_bundle_name] = item
+        self.potential_cache_cost += item.cost
+
+        while self.potential_cache_cost > self.potential_cache_max_total_cost:
             first_in = list(self.potential_txs.keys())[0]
+            self.potential_cache_max_total_cost -= self.potential_txs[first_in].cost
             self.potential_txs.pop(first_in)
 
     def get_spendbundle(self, bundle_hash: bytes32) -> Optional[SpendBundle]:
@@ -396,34 +462,38 @@ class MempoolManager:
             return self.mempool.spends[bundle_hash]
         return None
 
-    async def new_peak(self, new_peak: Optional[BlockRecord]) -> List[Tuple[SpendBundle, CostResult, bytes32]]:
+    async def new_peak(self, new_peak: Optional[BlockRecord]) -> List[Tuple[SpendBundle, NPCResult, bytes32]]:
         """
         Called when a new peak is available, we try to recreate a mempool for the new tip.
         """
         if new_peak is None:
             return []
+        if new_peak.is_transaction_block is False:
+            return []
         if self.peak == new_peak:
             return []
-        if new_peak.height <= self.constants.INITIAL_FREEZE_PERIOD:
+        assert new_peak.timestamp is not None
+        if new_peak.timestamp <= self.constants.INITIAL_FREEZE_END_TIMESTAMP:
             return []
 
         self.peak = new_peak
 
         old_pool = self.mempool
-        self.mempool = Mempool.create(self.mempool_size)
+        self.mempool = Mempool(self.mempool_max_total_cost)
 
         for item in old_pool.spends.values():
-            await self.add_spendbundle(item.spend_bundle, item.cost_result, item.spend_bundle_name, False)
+            await self.add_spendbundle(item.spend_bundle, item.npc_result, item.spend_bundle_name, False)
 
         potential_txs_copy = self.potential_txs.copy()
         self.potential_txs = {}
         txs_added = []
-        for tx, cached_result, cached_name in potential_txs_copy.values():
-            cost, status, error = await self.add_spendbundle(tx, cached_result, cached_name)
+        for item in potential_txs_copy.values():
+            cost, status, error = await self.add_spendbundle(item.spend_bundle, item.npc_result, item.spend_bundle_name)
             if status == MempoolInclusionStatus.SUCCESS:
-                txs_added.append((tx, cached_result, cached_name))
+                txs_added.append((item.spend_bundle, item.npc_result, item.spend_bundle_name))
         log.debug(
-            f"Size of mempool: {len(self.mempool.spends)}, minimum fee to get in: {self.mempool.get_min_fee_rate()}"
+            f"Size of mempool: {len(self.mempool.spends)} spends, cost: {self.mempool.total_mempool_cost} "
+            f"minimum fee to get in: {self.mempool.get_min_fee_rate(100000)}"
         )
         return txs_added
 
