@@ -4,7 +4,6 @@ import logging
 import socket
 import time
 import traceback
-from asyncio import Task
 from pathlib import Path
 from typing import Callable, Dict, List, Optional, Set, Tuple, Union
 
@@ -105,7 +104,7 @@ class WalletNode:
         self.backup_initialized = False  # Delay first launch sync after user imports backup info or decides to skip
         self.server = None
         self.wsm_close_task = None
-        self.sync_task: Optional[Task] = None
+        self.sync_task: Optional[asyncio.Task] = None
         self.new_peak_lock: Optional[asyncio.Lock] = None
         self.logged_in_fingerprint: Optional[int] = None
         self.peer_task = None
@@ -148,7 +147,10 @@ class WalletNode:
         path = path_from_root(self.root_path, db_path_replaced)
         mkdir(path.parent)
 
-        self.wallet_state_manager = await WalletStateManager.create(private_key, self.config, path, self.constants)
+        assert self.server is not None
+        self.wallet_state_manager = await WalletStateManager.create(
+            private_key, self.config, path, self.constants, self.server
+        )
 
         self.wsm_close_task = None
 
@@ -323,7 +325,7 @@ class WalletNode:
         if not self.has_full_node() and self.wallet_peers is not None:
             asyncio.create_task(self.wallet_peers.on_connect(peer))
 
-    async def _periodically_check_full_node(self):
+    async def _periodically_check_full_node(self) -> None:
         tries = 0
         while not self._shut_down and tries < 5:
             if self.has_full_node():
@@ -469,11 +471,14 @@ class WalletNode:
                 self.wallet_state_manager.sync_store.add_potential_peak(header_block)
                 self.start_sync()
 
-    def start_sync(self):
+    def start_sync(self) -> None:
         self.log.info("self.sync_event.set()")
         self.sync_event.set()
 
-    async def check_new_peak(self):
+    async def check_new_peak(self) -> None:
+        if self.wallet_state_manager is None:
+            return
+
         current_peak: Optional[BlockRecord] = self.wallet_state_manager.blockchain.get_peak()
         if current_peak is None:
             return
@@ -486,7 +491,7 @@ class WalletNode:
                 self.start_sync()
                 return
 
-    async def sync_job(self):
+    async def sync_job(self) -> None:
         while True:
             self.log.info("Loop start in sync job")
             if self._shut_down is True:
@@ -498,6 +503,7 @@ class WalletNode:
             if self._shut_down is True:
                 break
             try:
+                assert self.wallet_state_manager is not None
                 self.wallet_state_manager.set_sync_mode(True)
                 await self._sync()
             except Exception as e:
@@ -508,12 +514,12 @@ class WalletNode:
                     self.wallet_state_manager.set_sync_mode(False)
             self.log.info("Loop end in sync job")
 
-    async def _sync(self):
+    async def _sync(self) -> None:
         """
         Wallet has fallen far behind (or is starting up for the first time), and must be synced
         up to the LCA of the blockchain.
         """
-        if self.wallet_state_manager is None or self.backup_initialized is False:
+        if self.wallet_state_manager is None or self.backup_initialized is False or self.server is None:
             return
 
         highest_weight: uint128 = uint128(0)
@@ -543,25 +549,24 @@ class WalletNode:
             self.log.info("No peers to sync to")
             return
 
-        await self.wallet_state_manager.blockchain.lock.acquire()
-        try:
+        async with self.wallet_state_manager.blockchain.lock:
             fork_height = None
             if peak is not None:
                 fork_height = self.wallet_state_manager.sync_store.get_potential_fork_point(peak.header_hash)
             if fork_height is None:
-                fork_height = 0
+                fork_height = uint32(0)
             await self.wallet_state_manager.blockchain.warmup(fork_height)
             batch_size = self.constants.MAX_BLOCK_COUNT_PER_REQUESTS
             advanced_peak = False
             for i in range(max(0, fork_height - 1), peak_height, batch_size):
                 start_height = i
                 end_height = min(peak_height, start_height + batch_size)
-                peers: List[WSChiaConnection] = self.server.get_full_node_connections()
+                peers = self.server.get_full_node_connections()
                 added = False
                 for peer in peers:
                     try:
                         added, advanced_peak = await self.fetch_blocks_and_validate(
-                            peer, uint32(start_height), end_height, None if advanced_peak else fork_height
+                            peer, uint32(start_height), uint32(end_height), None if advanced_peak else fork_height
                         )
                         if added:
                             break
@@ -580,15 +585,13 @@ class WalletNode:
                         peak.height - self.constants.BLOCKS_CACHE_SIZE,
                     )
                 )
-        finally:
-            self.wallet_state_manager.blockchain.lock.release()
 
     async def fetch_blocks_and_validate(
         self,
         peer: WSChiaConnection,
         height_start: uint32,
         height_end: uint32,
-        fork_point_with_peak: uint32,
+        fork_point_with_peak: Optional[uint32],
     ) -> Tuple[bool, bool]:
         """
         Returns whether the blocks validated, and whether the peak was advanced
@@ -666,6 +669,8 @@ class WalletNode:
                 self.wallet_state_manager.state_changed("new_block")
             elif result == ReceiveBlockResult.INVALID_BLOCK:
                 raise ValueError("Value error peer sent us invalid block")
+        if advanced_peak:
+            await self.wallet_state_manager.create_more_puzzle_hashes()
         return True, advanced_peak
 
     def validate_additions(
