@@ -54,6 +54,8 @@ from chia.wallet.wallet_puzzle_store import WalletPuzzleStore
 from chia.wallet.wallet_sync_store import WalletSyncStore
 from chia.wallet.wallet_transaction_store import WalletTransactionStore
 from chia.wallet.wallet_user_store import WalletUserStore
+from chia.server.server import ChiaServer
+from chia.wallet.did_wallet.did_wallet import DIDWallet
 
 
 class WalletStateManager:
@@ -70,6 +72,8 @@ class WalletStateManager:
     # Makes sure only one asyncio thread is changing the blockchain state at one time
     lock: asyncio.Lock
 
+    tx_lock: asyncio.Lock
+
     log: logging.Logger
 
     # TODO Don't allow user to send tx until wallet is synced
@@ -95,6 +99,7 @@ class WalletStateManager:
     coin_store: WalletCoinStore
     sync_store: WalletSyncStore
     weight_proof_handler: Any
+    server: ChiaServer
 
     @staticmethod
     async def create(
@@ -102,20 +107,23 @@ class WalletStateManager:
         config: Dict,
         db_path: Path,
         constants: ConsensusConstants,
+        server: ChiaServer,
         name: str = None,
     ):
         self = WalletStateManager()
         self.new_wallet = False
         self.config = config
         self.constants = constants
+        self.server = server
 
         if name:
             self.log = logging.getLogger(name)
         else:
             self.log = logging.getLogger(__name__)
         self.lock = asyncio.Lock()
+        self.tx_lock = asyncio.Lock()
 
-        self.log.warning(f"Starting in db path: {db_path}")
+        self.log.debug(f"Starting in db path: {db_path}")
         self.db_connection = await aiosqlite.connect(db_path)
         self.db_wrapper = DBWrapper(self.db_connection)
         self.coin_store = await WalletCoinStore.create(self.db_wrapper)
@@ -147,27 +155,32 @@ class WalletStateManager:
         assert main_wallet_info is not None
 
         self.private_key = private_key
-
         self.main_wallet = await Wallet.create(self, main_wallet_info)
 
         self.wallets = {main_wallet_info.id: self.main_wallet}
 
+        wallet = None
         for wallet_info in await self.get_all_wallet_info_entries():
             # self.log.info(f"wallet_info {wallet_info}")
             if wallet_info.type == WalletType.STANDARD_WALLET:
                 if wallet_info.id == 1:
                     continue
                 wallet = await Wallet.create(config, wallet_info)
-                self.wallets[wallet_info.id] = wallet
             elif wallet_info.type == WalletType.COLOURED_COIN:
                 wallet = await CCWallet.create(
                     self,
                     self.main_wallet,
                     wallet_info,
                 )
-                self.wallets[wallet_info.id] = wallet
             elif wallet_info.type == WalletType.RATE_LIMITED:
                 wallet = await RLWallet.create(self, wallet_info)
+            elif wallet_info.type == WalletType.DISTRIBUTED_ID:
+                wallet = await DIDWallet.create(
+                    self,
+                    self.main_wallet,
+                    wallet_info,
+                )
+            if wallet is not None:
                 self.wallets[wallet_info.id] = wallet
 
         async with self.puzzle_store.lock:
@@ -209,6 +222,13 @@ class WalletStateManager:
                     wallet_info,
                 )
                 self.wallets[wallet_info.id] = wallet
+            elif wallet_info.type == WalletType.DISTRIBUTED_ID:
+                wallet = await DIDWallet.create(
+                    self,
+                    self.main_wallet,
+                    wallet_info,
+                )
+                self.wallets[wallet_info.id] = wallet
 
     async def get_keys(self, puzzle_hash: bytes32) -> Optional[Tuple[G1Element, PrivateKey]]:
         index_for_puzzlehash = await self.puzzle_store.index_for_puzzle_hash(puzzle_hash)
@@ -301,6 +321,33 @@ class WalletStateManager:
         if unused > 0:
             await self.puzzle_store.set_used_up_to(uint32(unused - 1))
 
+    async def update_wallet_puzzle_hashes(self, wallet_id):
+        derivation_paths: List[DerivationRecord] = []
+        target_wallet = self.wallets[wallet_id]
+        last: Optional[uint32] = await self.puzzle_store.get_last_derivation_path_for_wallet(wallet_id)
+        unused: Optional[uint32] = await self.puzzle_store.get_unused_derivation_path()
+        if unused is None:
+            # This handles the case where the database has entries but they have all been used
+            unused = await self.puzzle_store.get_last_derivation_path()
+            if unused is None:
+                # This handles the case where the database is empty
+                unused = uint32(0)
+        for index in range(unused, last):
+            pubkey: G1Element = self.get_public_key(uint32(index))
+            puzzle: Program = target_wallet.puzzle_for_pk(bytes(pubkey))
+            puzzlehash: bytes32 = puzzle.get_tree_hash()
+            self.log.info(f"Generating public key at index {index} puzzle hash {puzzlehash.hex()}")
+            derivation_paths.append(
+                DerivationRecord(
+                    uint32(index),
+                    puzzlehash,
+                    pubkey,
+                    target_wallet.wallet_info.type,
+                    uint32(target_wallet.wallet_info.id),
+                )
+            )
+        await self.puzzle_store.add_derivation_paths(derivation_paths)
+
     async def get_unused_derivation_record(self, wallet_id: uint32) -> DerivationRecord:
         """
         Creates a puzzle hash for the given wallet, and then makes more puzzle hashes
@@ -368,7 +415,7 @@ class WalletStateManager:
             return
         self.state_changed_callback(state, wallet_id, data_object)
 
-    def tx_pending_changed(self):
+    def tx_pending_changed(self) -> None:
         """
         Notifies the wallet node that there's new tx pending
         """
@@ -533,6 +580,17 @@ class WalletStateManager:
             if prev.is_transaction_block:
                 break
             prev = await self.blockchain.get_block_record_from_db(prev.prev_hash)
+        wallet_ids: Set[int] = set()
+        for coin in coins:
+            info = await self.puzzle_store.wallet_info_for_puzzle_hash(coin.puzzle_hash)
+            if info is not None:
+                wallet_ids.add(info[0])
+
+        all_outgoing_tx: Dict[int, List[TransactionRecord]] = {}
+        for wallet_id in wallet_ids:
+            all_outgoing_tx[wallet_id] = await self.tx_store.get_all_transactions(
+                wallet_id, TransactionType.OUTGOING_TX
+            )
 
         for coin in coins:
             if coin.name() in trade_additions:
@@ -548,7 +606,12 @@ class WalletStateManager:
             info = await self.puzzle_store.wallet_info_for_puzzle_hash(coin.puzzle_hash)
             if info is not None:
                 wallet_id, wallet_type = info
-                await self.coin_added(coin, is_coinbase, is_fee_reward, uint32(wallet_id), wallet_type, height)
+                await self.coin_added(
+                    coin, is_coinbase, is_fee_reward, uint32(wallet_id), wallet_type, height, all_outgoing_tx[wallet_id]
+                )
+            derivation_index = await self.puzzle_store.index_for_puzzle_hash(coin.puzzle_hash)
+            if derivation_index is not None:
+                await self.puzzle_store.set_used_up_to(derivation_index, True)
 
         return trade_adds
 
@@ -562,33 +625,23 @@ class WalletStateManager:
         # Keep track of trade coins that are removed
         trade_coin_removed: List[Coin] = []
 
+        all_unconfirmed: List[TransactionRecord] = await self.tx_store.get_all_unconfirmed()
         for coin in coins:
-            self.log.info(f"Coin removed: {coin.name()}")
             record = await self.coin_store.get_coin_record_by_coin_id(coin.name())
             if coin.name() in trade_removals:
-                self.log.info(f"Coin:{coin.name()} is part of trade")
                 trade_coin_removed.append(coin)
             if record is None:
-                self.log.info(f"Coin:{coin.name()} NO RECORD")
                 continue
-            self.log.info(f"Coin:{coin.name()} Setting removed")
-            await self.coin_removed(coin, height, record.wallet_id)
+            await self.coin_store.set_spent(coin.name(), height)
+            for unconfirmed_record in all_unconfirmed:
+                for rem_coin in unconfirmed_record.removals:
+                    if rem_coin.name() == coin.name():
+                        await self.tx_store.set_confirmed(unconfirmed_record.name, height)
+
+            self.state_changed("coin_removed", record.wallet_id)
 
         return trade_coin_removed
 
-    async def coin_removed(self, coin: Coin, height: uint32, wallet_id: int):
-        """
-        Called when coin gets spent
-        """
-
-        await self.coin_store.set_spent(coin.name(), height)
-
-        unconfirmed_record: List[TransactionRecord] = await self.tx_store.unconfirmed_with_removal_coin(coin.name())
-        for unconfirmed in unconfirmed_record:
-            await self.tx_store.set_confirmed(unconfirmed.name, height)
-
-        self.state_changed("coin_removed", wallet_id)
-
     async def coin_added(
         self,
         coin: Coin,
@@ -597,11 +650,11 @@ class WalletStateManager:
         wallet_id: uint32,
         wallet_type: WalletType,
         height: uint32,
+        all_outgoing_transaction_records: List[TransactionRecord],
     ):
         """
         Adding coin to DB
         """
-        self.log.info(f"Adding coin: {coin} at {height}")
         farm_reward = False
         if coinbase or fee_reward:
             farm_reward = True
@@ -629,7 +682,11 @@ class WalletStateManager:
             )
             await self.tx_store.add_transaction_record(tx_record, True)
         else:
-            records = await self.tx_store.tx_with_addition_coin(coin.name(), wallet_id)
+            records: List[TransactionRecord] = []
+            for record in all_outgoing_transaction_records:
+                for add_coin in record.additions:
+                    if add_coin.name() == coin.name():
+                        records.append(record)
 
             if len(records) > 0:
                 # This is the change from this transaction
@@ -663,8 +720,8 @@ class WalletStateManager:
         )
         await self.coin_store.add_coin_record(coin_record)
 
-        if wallet_type == WalletType.COLOURED_COIN:
-            wallet: CCWallet = self.wallets[wallet_id]
+        if wallet_type == WalletType.COLOURED_COIN or wallet_type == WalletType.DISTRIBUTED_ID:
+            wallet = self.wallets[wallet_id]
             header_hash: bytes32 = self.blockchain.height_to_hash(height)
             block: Optional[HeaderBlockRecord] = await self.block_store.get_header_block_record(header_hash)
             assert block is not None
@@ -677,7 +734,7 @@ class WalletStateManager:
         """
         Called from wallet before new transaction is sent to the full_node
         """
-        if self.peak is None or self.peak.height <= self.constants.INITIAL_FREEZE_PERIOD:
+        if self.peak is None or int(time.time()) <= self.constants.INITIAL_FREEZE_END_TIMESTAMP:
             raise ValueError("Initial Freeze Period")
         # Wallet node will use this queue to retry sending this transaction until full nodes receives it
         await self.tx_store.add_transaction_record(tx_record, False)
@@ -754,28 +811,28 @@ class WalletStateManager:
             if coin.confirmed_block_height <= fork_h:
                 unspent_coin_names.add(coin.name())
 
-        # # Get all blocks after fork point up to but not including this block
-        # curr: BlockRecord = self.lockchain.blocks[new_block.prev_header_hash]
-        # reorg_blocks: List[HeaderBlockRecord] = []
-        # while curr.height > fork_h:
-        #     header_block_record = await self.block_store.get_header_block_record(
-        #         curr.header_hash
-        #     )
-        #     reorg_blocks.append(header_block_record)
-        #     curr = self.blockchain.blocks[curr.prev_header_hash]
-        # reorg_blocks.reverse()
+        # Get all blocks after fork point up to but not including this block
+        if new_block.height > 0:
+            curr: BlockRecord = self.blockchain.block_record(new_block.prev_hash)
+            reorg_blocks: List[HeaderBlockRecord] = []
+            while curr.height > fork_h:
+                header_block_record = await self.block_store.get_header_block_record(curr.header_hash)
+                assert header_block_record is not None
+                reorg_blocks.append(header_block_record)
+                if curr.height == 0:
+                    break
+                curr = self.blockchain.block_record(curr.prev_hash)
+            reorg_blocks.reverse()
 
-        # For each block, process additions to get all Coins, then process removals to get unspent coins
-        # for reorg_block in reorg_blocks:
-        #     for addition in reorg_block.additions:
-        #         unspent_coin_names.add(addition.name())
-        #     for removal in reorg_block.removals:
-        #         record = await self.puzzle_store.get_derivation_record_for_puzzle_hash(
-        #             removal.puzzle_hash
-        #         )
-        #         if record is None:
-        #             continue
-        #         unspent_coin_names.remove(removal)
+            # For each block, process additions to get all Coins, then process removals to get unspent coins
+            for reorg_block in reorg_blocks:
+                for addition in reorg_block.additions:
+                    unspent_coin_names.add(addition.name())
+                for removal in reorg_block.removals:
+                    record = await self.puzzle_store.get_derivation_record_for_puzzle_hash(removal.puzzle_hash)
+                    if record is None:
+                        continue
+                    unspent_coin_names.remove(removal)
 
         my_puzzle_hashes = self.puzzle_store.all_puzzle_hashes
 
@@ -871,11 +928,11 @@ class WalletStateManager:
                 TransactionType.OUTGOING_TRADE,
                 TransactionType.INCOMING_TRADE,
             ]:
-                await self.tx_store.tx_reorged(record.name())
+                await self.tx_store.tx_reorged(record.name)
 
         self.tx_pending_changed()
 
-    async def close_all_stores(self):
+    async def close_all_stores(self) -> None:
         if self.blockchain is not None:
             self.blockchain.shut_down()
         await self.db_connection.close()
@@ -942,7 +999,7 @@ class WalletStateManager:
         backup_file_text = json.dumps(backup)
         file_path.write_text(backup_file_text)
 
-    async def import_backup_info(self, file_path):
+    async def import_backup_info(self, file_path) -> None:
         json_dict = open_backup_file(file_path, self.private_key)
         wallet_list_json = json_dict["data"]["wallet_list"]
 
@@ -969,6 +1026,29 @@ class WalletStateManager:
         self.wallets[uint32(wallet_id)] = wallet
         await self.create_more_puzzle_hashes()
 
+    # search through the blockrecords and return the most recent coin to use a given puzzlehash
+    async def search_blockrecords_for_puzzlehash(self, puzzlehash: bytes32):
+        header_hash_of_interest = None
+        heighest_block_height = 0
+        peak: Optional[BlockRecord] = self.blockchain.get_peak()
+        if peak is None:
+            return None, None
+        peak_block: Optional[HeaderBlockRecord] = await self.blockchain.block_store.get_header_block_record(
+            peak.header_hash
+        )
+        while peak_block is not None:
+            tx_filter = PyBIP158([b for b in peak_block.header.transactions_filter])
+            if tx_filter.Match(bytearray(puzzlehash)) and peak_block.height > heighest_block_height:
+                header_hash_of_interest = peak_block.header_hash
+                heighest_block_height = peak_block.height
+                break
+            else:
+                peak_block = await self.blockchain.block_store.get_header_block_record(
+                    peak_block.header.prev_header_hash
+                )
+
+        return heighest_block_height, header_hash_of_interest
+
     async def get_spendable_coins_for_wallet(self, wallet_id: int, records=None) -> Set[WalletCoinRecord]:
         if self.peak is None:
             return set()
