@@ -10,7 +10,7 @@ wallet_program_files = set(
         "chia/wallet/puzzles/calculate_synthetic_public_key.clvm",
         "chia/wallet/puzzles/cc.clvm",
         "chia/wallet/puzzles/chialisp_deserialisation.clvm",
-        # "chia/wallet/puzzles/generator.clvm", - see chia/wallet/puzzles/lowlevel_generator.py
+        "chia/wallet/puzzles/rom_bootstrap_generator.clvm",
         "chia/wallet/puzzles/generator_for_single_coin.clvm",
         "chia/wallet/puzzles/genesis-by-coin-id-with-0.clvm",
         "chia/wallet/puzzles/genesis-by-puzzle-hash-with-0.clvm",
@@ -24,6 +24,14 @@ wallet_program_files = set(
         "chia/wallet/puzzles/rl_aggregation.clvm",
         "chia/wallet/puzzles/rl.clvm",
         "chia/wallet/puzzles/sha256tree_module.clvm",
+        "chia/wallet/puzzles/singleton_top_layer.clvm",
+        "chia/wallet/puzzles/did_innerpuz.clvm",
+        "chia/wallet/puzzles/decompress_puzzle.clvm",
+        "chia/wallet/puzzles/decompress_coin_solution_entry_with_prefix.clvm",
+        "chia/wallet/puzzles/decompress_coin_solution_entry.clvm",
+        "chia/wallet/puzzles/block_program_zero.clvm",
+        "chia/wallet/puzzles/test_generator_deserialize.clvm",
+        "chia/wallet/puzzles/test_multiple_generator_input_arguments.clvm",
     ]
 )
 
